/*
** Copyright (c) 1996, 2002
**      Jeff Forys (jeff@forys.cranbury.nj.us).  All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that: (1) Redistributions of
** source code must retain the above copyright notice, this list of
** conditions and the following disclaimer, (2) Redistributions in
** binary form must reproduce the above copyright notice, this list
** of conditions and the following disclaimer in the documentation
** and/or other materials provided with the distribution, (3) All
** advertising materials mentioning features or use of this software
** must display the following acknowledgment: ``This product includes
** software developed by Jeff Forys (jeff@forys.cranbury.nj.us).'', (4)
** The name of the author may not be used to endorse or promote products
** derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
** WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
** MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*/

#ifndef lint
static char rcsid[] = "$Id: ifops.c,v 1.10 2004/01/24 04:36:57 forys Exp $";
#endif

#include <sys/param.h>
#include <sys/types.h>
#include <sys/wait.h>

#include <sys/ioctl.h>
#include <sys/socket.h>
#ifdef SVR4
#include <sys/sockio.h>
#endif
#include <net/if.h>

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <fcntl.h>
#include <errno.h>

#include <X11/IntrinsicP.h>

#include "NetintfP.h"

extern char *ProgName;

#define	IF_MAXARGS	256		/* maximum number of arguments */
#define	IF_MAXLINE	8192		/* maximum size of file */
#define	IF_FILPRFX	"/.if"		/* prefix for interface config file */

/*
 * List of states an interface can be in.
 */
typedef enum {
	IFCMD_UP,			/* bring interface up */
	IFCMD_INIT,			/* check if interface is coming up */
	IFCMD_UPEX,			/* exec after interface is up */
	IFCMD_DOWN,			/* bring interface down */
	IFCMD_DNEX			/* exec after interface is down */
} NetintfCmds;

/*
 * Return codes from ExecCmd()
 */
typedef enum {
	EXEC_ERROR,			/* could not execute command */
	EXEC_STATOK,			/* command executed, returned 0 */
	EXEC_STATERR,			/* command executed, returned non-0 */
	EXEC_NULLCMD			/* NULL command in argvec */
} ExecRtn;

/*
 * ReadCmd(cmd, path, intf)
 *
 * Read command (cmd) from (path)(IF_FILPRFX)(intf) and parse into static
 * list of arguments (suitable for execve(2)).  Lines that start with a
 * '#' are ignored.  Return 'char *av[]' on success, NULL on failure.
 *
 * Notes:
 *   - The maximum size of the file is IF_MAXLINE bytes.
 *   - Potentially, we can return an av[] with a single NULL argument;
 *     ExecCmd() will deal with that.
 */
static const char **
ReadCmd(cmd, path, intf)
	NetintfCmds cmd;
	const char *path;
	const char *intf;
{
	static const char *av[IF_MAXARGS + 1];
	static char readln[IF_MAXLINE + 1];
	char *scp, *ecp;
	const char *cp;
	int nbytes, skipunit = 0;

again:
	/*
	 * Mash together (path)(IF_FILPRFX)(intf) into "readln", taking
	 * care as to not overrun the buffer.
	 */
	for (scp = &readln[0], ecp = &readln[IF_MAXLINE], cp = path;
	     scp < ecp && *cp; scp++, cp++)
		*scp = *cp;
	for (cp = IF_FILPRFX; scp < ecp && *cp; scp++, cp++)
		*scp = *cp;
	for (cp = intf; scp < ecp && *cp; scp++, cp++) {
		if (isdigit(*cp) && skipunit)
			break;
		*scp = *cp;
	}
	*scp = '\0';

	/*
	 * Open the file, read up to IF_MAXLINE bytes and close it.
	 * If an error occurred during open() or read(), return NULL.
	 */
	{
		int fd;

		if ((fd = open(readln, O_RDONLY, 0)) < 0) {
			if (skipunit)
				return NULL;
			skipunit = 1;
			goto again;
		}
		nbytes = read(fd, readln, IF_MAXLINE);
		close(fd);

		if (nbytes < 0)
			return NULL;
	}

	/*
	 * Determine what line we are interested in, and advance 'scp' to
	 * the beginning of it (skipping over any comment lines).
	 */
	{
		int newline = 1;
		int linecnt = 0;
		int lineno;

		if ((lineno = (cmd == IFCMD_UP)?   1: (cmd == IFCMD_INIT)? 2:
		              (cmd == IFCMD_UPEX)? 3: (cmd == IFCMD_DOWN)? 4:
		              (cmd == IFCMD_DNEX)? 5: 0) == 0)
			return NULL;

		for (scp = &readln[0]; nbytes; scp++, nbytes--) {
			if (newline == 1 && *scp == '#') {
				newline = 2;
			} else if (newline == 2) {
				if (*scp == '\n')
					newline = 1;
			} else {
				if (newline == 1 && ++linecnt == lineno)
					break;

				newline = (*scp == '\n')? 1: 0;
			}
		}

		if (linecnt != lineno)	/* not enough lines in file */
			return NULL;

		ecp = scp + nbytes;
	}

	/*
	 * Parse line (starting at scp) into argument vector (*av[]).
	 * We can not fail now.
	 */
	{
		int i, inarg = 0;

		for (i = 0; scp < ecp; scp++) {
			if (*scp == '\n') {
				*scp = '\0';
				break;
			} else if (inarg && isspace(*scp)) {
				inarg = 0;
				*scp = '\0';
			} else if (!inarg && !isspace(*scp)) {
				inarg = 1;
				if (i < IF_MAXARGS)
					av[i++] = scp;
			}
		}

		av[i] = NULL;
	}

	return(av);
}

/*
 * ExecCmd(av)
 *
 * Fork a child and exec *av[] (with current environment variables).
 * Return ExecRtn indication.
 */
static ExecRtn
ExecCmd(av)
	const char * av[];
{
	extern char **environ;
	pid_t pid, wpid;
	int s, rval = EXEC_ERROR;

	/*
	 * If av[0] is NULL, assume nothing was desired.
	 */
	if (av[0] == NULL)
		return EXEC_NULLCMD;

	/*
	 * If command is executable, fork and exec (while parent waits...).
	 */
	if (access(av[0], X_OK) == 0) {
		switch(pid = fork()) {
		    case -1:	/* error */
			break;
		    case 0:	/* child */
			/*
			 * Close parent's stdin/stdout/stderr and open
			 * "/dev/null" as their replacement.  The only
			 * thing we want to see is the exit status...
			 */
			(void) close(STDIN_FILENO);
			(void) close(STDOUT_FILENO);
			(void) close(STDERR_FILENO);
			(void) open("/dev/null", O_RDONLY);
			(void) open("/dev/null", O_WRONLY);
			dup(1);
			execve(av[0], (char * const *)av, environ);
			exit(1);
			break;
		    default:	/* parent */
			for (;;) {
				if ((wpid = waitpid(pid, &s, 0)) < 0 &&
				    errno != EINTR)
					break;
				else if (WIFEXITED(s)) {
					rval = (WEXITSTATUS(s) == 0)?
						EXEC_STATOK: EXEC_STATERR;
					break;
				} else if (WIFSIGNALED(s))
					break;
			}
			break;
		}
	}

	return rval;
}

/*
 * netiferr(w)
 *
 * Set error condition for NetintfWidget's interface.
 */
static void
netiferr(w)
	NetintfWidget w;
{
	if (w->netintf.sockfd >= 0)
		(void) close(w->netintf.sockfd);
	w->netintf.sockfd = -1;
	SETSTATE(w,IFST_ERR);
}

#define	DBGREAD(w, av)							\
if (w->netintf.dbglvl > 0) {						\
	const char **avidx;						\
	fprintf(stderr, "%s: DEBUG: ReadCmd: ", ProgName);		\
	if (av == NULL) {						\
		fputs("NULL\n", stderr);				\
	} else {							\
		for (avidx = av; *avidx != NULL; avidx++)		\
			fprintf(stderr, "<%s> ", *avidx);		\
		fputs("\n", stderr);					\
	}								\
}
#define	DBGEXEC(w, av, exret)						\
if (w->netintf.dbglvl > 0 && av != NULL) {				\
	fprintf(stderr, "%s: DEBUG: ExecCmd: %s\n", ProgName,		\
	        exret == EXEC_ERROR? "exec error":			\
	        exret == EXEC_STATOK? "exec ok, exit ok":		\
	        exret == EXEC_STATERR? "exec ok, exit error": "exec null"); \
}

/*
 * NetIfUp(w)
 *
 * Execute command to bring interface "w" up.
 */
void
NetIfUp(w)
	NetintfWidget w;
{
	const char **av;
	ExecRtn exret = EXEC_ERROR;	/* unnecessary initialization (GCC) */

	av = ReadCmd(IFCMD_UP, w->netintf.confdir, w->netintf.ifname);
	DBGREAD(w, av);
	if (av == NULL ||
	    (exret = ExecCmd(av)) == EXEC_ERROR || exret == EXEC_STATERR)
		netiferr(w);
	DBGEXEC(w, av, exret);
}

/*
 * NetIfDn(w)
 *
 * Execute command to bring interface "w" down.
 */
void
NetIfDn(w)
	NetintfWidget w;
{
	const char **av;
	ExecRtn exret = EXEC_ERROR;	/* unnecessary initialization (GCC) */

	av = ReadCmd(IFCMD_DOWN, w->netintf.confdir, w->netintf.ifname);
	DBGREAD(w, av);
	if (av == NULL ||
	    (exret = ExecCmd(av)) == EXEC_ERROR || exret == EXEC_STATERR)
		netiferr(w);
	DBGEXEC(w, av, exret);
}

/*
 * NetIfIsInit(w)
 *
 * Return TRUE if interface "w" is coming up, FALSE otherwise.
 */
Boolean
NetIfIsInit(w)
	NetintfWidget w;
{
	const char **av = NULL;		/* unnecessary initialization (GCC) */
	Boolean rval = TRUE;
	ExecRtn exret = EXEC_ERROR;	/* unnecessary initialization (GCC) */

	/*
	 * 1) Do not waste time checking interface in IFST_DOWN state;
	 *    if interface is on it's way up, it wont be in IFST_DOWN.
	 * 2) Only accept EXEC_STATOK (i.e. not EXEC_NULLCMD) as an
	 *    indication that we are coming up.
	 * 3) If the config file does not specify a way to determine if
	 *    initialization is underway (EXEC_NULLCMD) and state is
	 *    IFST_TOINIT or IFST_INIT, return TRUE; user is resposible
	 *    for resetting state to IFST_DOWN on failure.
	 */
	if (GETSTATE(w) == IFST_DOWN ||
	    (av = ReadCmd(IFCMD_INIT, w->netintf.confdir, w->netintf.ifname))
	        == NULL ||
	    (exret = ExecCmd(av)) != EXEC_STATOK) {
		if (exret == EXEC_NULLCMD &&
		    (GETSTATE(w) == IFST_TOINIT || GETSTATE(w) == IFST_INIT))
			rval = TRUE;
		else
			rval = FALSE;
	}

	if (GETSTATE(w) != IFST_DOWN) {
		DBGREAD(w, av);
		DBGEXEC(w, av, exret);
	}
	return rval;
}

/*
 * NetIfExecUp(w)
 *
 * After interface "w" comes up, execute a command.
 */
void
NetIfExecUp(w)
	NetintfWidget w;
{
	const char **av;
	ExecRtn exret = EXEC_ERROR;	/* unnecessary initialization (GCC) */

	av = ReadCmd(IFCMD_UPEX, w->netintf.confdir, w->netintf.ifname);
	DBGREAD(w, av);
	if (av != NULL)
		exret = ExecCmd(av);
	DBGEXEC(w, av, exret);
}

/*
 * NetIfExecDn(w)
 *
 * After interface "w" goes down, execute a command.
 */
void
NetIfExecDn(w)
	NetintfWidget w;
{
	const char **av;
	ExecRtn exret = EXEC_ERROR;	/* unnecessary initialization (GCC) */

	av = ReadCmd(IFCMD_DNEX, w->netintf.confdir, w->netintf.ifname);
	DBGREAD(w, av);
	if (av != NULL)
		exret = ExecCmd(av);
	DBGEXEC(w, av, exret);
}

/*
 * NetIfState(w)
 *
 * Return current interface state.
 */
NetintfState
NetIfState(w)
	NetintfWidget w;
{
	struct ifreq ifr;
	NetintfState ifstate = IFST_DOWN;

	/*
	 * Create a socket and if successful, use it to query the interface.
	 */
	if (w->netintf.sockfd < 0)
		if ((w->netintf.sockfd = socket(PF_INET, SOCK_DGRAM, 0)) < 0)
			ifstate = IFST_ERR;

	if (ifstate != IFST_ERR) {
		strncpy(ifr.ifr_name, w->netintf.ifname, IFNAMSIZ);
		if (ioctl(w->netintf.sockfd, SIOCGIFFLAGS, &ifr) < 0) {
			if (w->netintf.dbglvl > 0)
				fprintf(stderr, "%s: DEBUG: ioctl(SIOCGIFFLAGS) ret'd %d\n",
				        ProgName, errno);
			if (w->netintf.noif == 0) {
				(void) close(w->netintf.sockfd);
				w->netintf.sockfd = -1;
				ifstate = IFST_ERR;
			}
		} else {
			if (w->netintf.dbglvl > 1)
				fprintf(stderr, "%s: DEBUG: ifflags 0x%x\n",
				        ProgName, ifr.ifr_flags);
			if (ifr.ifr_flags & IFF_UP)
				ifstate = IFST_UP;
			else if (ifr.ifr_flags & IFF_RUNNING)
				ifstate = IFST_RUN;
			else if (NetIfIsInit(w))
				ifstate = IFST_INIT;
		}
	}

	return ifstate;
}
